"""Contains all the Mastertherm Exception classes."""


class MasterthermError(Exception):
    """Base Exception for all Mastertherm Exceptions"""

    def __init__(self, status, message):
        """Initialize."""
        super().__init__(status)
        self.status = status
        self.message = message


class MasterthermUnsupportedVersion(MasterthermError):
    """Raised when a version that is not supported is used."""


class MasterthermAuthenticationError(MasterthermError):
    """Raised when login returns wrong result."""


class MasterthermConnectionError(MasterthermError):
    """Raised when communication ended in error."""


class MasterthermResponseFormatError(MasterthermError):
    """Raised when page returns an unexpected response."""


class MasterthermTokenInvalid(MasterthermError):
    """Raised when page returns the Token is not valie."""


class MasterthermUnsupportedRole(MasterthermError):
    """Raised when connecting to an Unsupported Role."""


class MasterthermEntryNotFound(MasterthermError):
    """Raised when the data entry update requested is not found."""



"""Mastertherm API Client Class, handle integration."""
import logging
import time

from datetime import datetime, timedelta
from hashlib import sha1
from json.decoder import JSONDecodeError
from urllib.parse import urljoin, quote_plus

from aiohttp import ClientSession, ClientConnectionError, ContentTypeError

from .const import (
    APP_CLIENTINFO,
    APP_CLIENTINFO_NEW,
    DATE_FORMAT,
    SUPPORTED_ROLES,
    SUPPORTED_API_VERSIONS,
    URL_BASE,
    URL_BASE_NEW,
    URL_LOGIN,
    URL_LOGIN_NEW,
    URL_MODULES_NEW,
    URL_PUMPDATA,
    URL_PUMPDATA_NEW,
    URL_PUMPINFO,
    URL_PUMPINFO_NEW,
    URL_POSTUPDATE,
    URL_POSTUPDATE_NEW,
)
from .exceptions import (
    MasterthermAuthenticationError,
    MasterthermConnectionError,
    MasterthermUnsupportedRole,
    MasterthermUnsupportedVersion,
    MasterthermTokenInvalid,
    MasterthermResponseFormatError,
)

_LOGGER: logging.Logger = logging.getLogger(__package__)


class MasterthermAPI:
    """API Handler for the Mastertherm API."""

    def __init__(
        self,
        username: str,
        password: str,
        session: ClientSession,
        api_version: str,
    ) -> None:
        """Initialise the Mastertherm API Client.

        Parameters:
            username (str): The Login Username
            password (str): The Login Password
            session (ClientSession): an aiohttp client session
            api_version (str): The version of the API, mainly the host
                "v1"  : Original version, data response in varfile_mt1_config1 or 2
                "v2"  : New version since 2022 response in varFileData

        Return:
            The MasterthermAPI object

        Raises:
            MasterthermUnsupportedVersion: API Version is not supported."""
        if api_version not in SUPPORTED_API_VERSIONS:
            raise MasterthermUnsupportedVersion(
                "-1", f"Unsupported Version {api_version}"
            )

        self.__session = session
        self.__api_version = api_version
        self.__token = None
        self.__expires = None

        # Setup the Session Details based on if Old or New API.
        codeduser = quote_plus(username)
        if self.__api_version == "v1":
            hashpass = sha1(password.encode("utf-8")).hexdigest()
            self.__login_params = (
                f"login=login&uname={codeduser}&upwd={hashpass}&{APP_CLIENTINFO}"
            )
        else:
            codedpass = quote_plus(password)
            self.__login_params = (
                f"grant_type=password&username={codeduser}&"
                + f"password={codedpass}&{APP_CLIENTINFO_NEW}"
            )

    async def __token_expired(self) -> bool:
        """Return if the token has expired."""
        if self.__expires is None:
            return True

        if self.__api_version == "v1":
            if self.__expires <= datetime.fromtimestamp(time.mktime(time.gmtime())):
                return True
        else:
            if self.__expires <= datetime.now():
                return True

        if self.__expires <= datetime.now():
            return True

        return False

    async def __post(self, url: str, params: str) -> dict:
        """Push updates to the API."""
        if await self.__token_expired():
            await self.connect()

        try:
            if self.__api_version == "v1":
                # Original uses post, with Cookie Token
                response = await self.__session.post(
                    urljoin(URL_BASE, url),
                    data=params,
                    headers={"Content-Type": "application/x-www-form-urlencoded"},
                    cookies={"PHPSESSID": self.__token, "$version": "1"},
                )
            else:
                # New uses get, with Authorization Bearer
                response = await self.__session.post(
                    urljoin(URL_BASE_NEW, url),
                    data=params,
                    headers={
                        "Authorization": f"Bearer {self.__token}",
                        "Content-Type": "application/x-www-form-urlencoded",
                        "Host": "mastertherm.online",
                        "Connection": "close",
                    },
                )
        except ClientConnectionError as ex:
            _LOGGER.error("Client Connection Error: %s", ex)
            raise MasterthermConnectionError("3", "Client Connection Error") from ex
        except JSONDecodeError as ex:
            _LOGGER.error("JSON Decode Error: %s", ex)
            raise MasterthermConnectionError("3", "JSON Decode Error") from ex

        # Version 2 responds with an error and json.
        # We should only get something other than 200 if the servers are down.
        if response.status != 200:
            try:
                response_json = await response.json()

                # Deal with the v2 error if we have a response
                if response_json["status"]["id"] == 401:
                    raise MasterthermTokenInvalid("1", response_json)
                else:
                    _LOGGER.error("Mastertherm API some other error: %s", response_json)
                    raise MasterthermResponseFormatError("2", response_json)
            except ContentTypeError as ex:
                response_text = await response.text()
                _LOGGER.error(
                    "Mastertherm API Connection Error %s:%s",
                    str(response.status),
                    response_text,
                )
                raise MasterthermConnectionError(
                    str(response.status), response_text
                ) from ex

        # Convert to JSON if possible and return the result
        try:
            response_json = await response.json()
        except ContentTypeError as exc:
            response_text = await response.text()
            if response_text == "User not logged in":

                raise MasterthermTokenInvalid("1", response_text) from exc
            else:
                _LOGGER.error("Mastertherm API some other error: %s", response_text)
                raise MasterthermResponseFormatError("2", response_text) from exc

        return response_json

    async def __get(self, url: str, params: str) -> dict:
        """Get updates from the API, for old this mostly uses Post."""
        if await self.__token_expired():
            await self.connect()

        try:
            if self.__api_version == "v1":
                # Original uses post, with Cookie Token
                response = await self.__session.post(
                    urljoin(URL_BASE, url),
                    data=params,
                    headers={"content-type": "application/x-www-form-urlencoded"},
                    cookies={"PHPSESSID": self.__token, "$version": "1"},
                )
            else:
                # New uses get, with Authorization Bearer
                response = await self.__session.get(
                    urljoin(URL_BASE_NEW, url),
                    params=params,
                    headers={
                        "Authorization": f"Bearer {self.__token}",
                        "Host": "mastertherm.online",
                        "Connection": "close",
                    },
                )
        except ClientConnectionError as ex:
            _LOGGER.error("Client Connection Error: %s", ex)
            raise MasterthermConnectionError("3", "Client Connection Error") from ex
        except JSONDecodeError as ex:
            _LOGGER.error("JSON Decode Error: %s", ex)
            raise MasterthermConnectionError("3", "JSON Decode Error") from ex

        # Version 2 responds with an error and json.
        # We should only get something other than 200 if the servers are down.
        if response.status != 200:
            try:
                response_json = await response.json()

                # Deal with the v2 error if we have a response
                if response_json["status"]["id"] == 401:
                    raise MasterthermTokenInvalid("1", response_json)
                else:
                    _LOGGER.error("Mastertherm API some other error: %s", response_json)
                    raise MasterthermResponseFormatError("2", response_json)
            except ContentTypeError as ex:
                response_text = await response.text()
                _LOGGER.error(
                    "Mastertherm API Connection Error %s:%s",
                    str(response.status),
                    response_text,
                )
                raise MasterthermConnectionError(
                    str(response.status), response_text
                ) from ex

        # Convert to JSON if possible and return the result
        try:
            response_json = await response.json()
        except ContentTypeError as exc:
            response_text = await response.text()
            if response_text == "User not logged in":

                raise MasterthermTokenInvalid("1", response_text) from exc
            else:
                _LOGGER.error("Mastertherm API some other error: %s", response_text)
                raise MasterthermResponseFormatError("2", response_text) from exc

        return response_json

    def get_url(self) -> str:
        """Return the API URL Used.

        Returns:
            URL(str): The API URL for the version."""
        if self.__api_version == "v1":
            return URL_BASE
        else:
            return URL_BASE_NEW

    async def connect(self) -> dict:
        """Perform the connection to the Mastertherm API Server:

        Returns:
             devices (dict): Return the list of devices, modules and units.

        Raises:
            MasterthermConnectionError - Failed to Connect
            MasterthermAuthenticationError - Failed to Authenticate
            MasterthermUnsupportedRole - Role is not in supported roles"""
        # Connect based on requirements
        if self.__api_version == "v1":
            # Clear out cookies, clears the auth token.
            url = urljoin(URL_BASE, URL_LOGIN)
        else:
            url = urljoin(URL_BASE_NEW, URL_LOGIN_NEW)

        try:
            response = await self.__session.post(
                url,
                data=self.__login_params,
                cookies={"PHPSESSID": "", "$version": "1"},
                headers={"content-type": "application/x-www-form-urlencoded"},
            )
        except ClientConnectionError as ex:
            _LOGGER.error("Client Connection Error: %s", ex)
            raise MasterthermConnectionError("3", "Client Connection Error") from ex

        # Response shoudl always be 200 even for login failures
        if response.status != 200:
            error_msg = await response.text()
            raise MasterthermConnectionError(str(response.status), error_msg)

        response_json = await response.json()

        # Separate process for Old and New API's
        if self.__api_version == "v1":
            # Old Process gets return and modules in one go
            # token and expiry is stored in a cookie.
            if response_json["returncode"] != 0:
                _LOGGER.error("Authentication Error: %s", response_json)
                raise MasterthermAuthenticationError(
                    response_json["returncode"], response_json["message"]
                )

            # Get or Refresh the Token and Expiry
            self.__token = response.cookies["PHPSESSID"].value
            self.__expires = datetime.strptime(
                response.cookies["PHPSESSID"]["expires"], DATE_FORMAT
            )
        else:
            # New process uses JSON responses with Openconnect ID.
            # Requires an additional call to get the modules.
            if "error" in response_json:
                raise MasterthermAuthenticationError(
                    response_json["error"], response_json["error_description"]
                )

            #  Get or Refresh the Token and Expiry
            self.__token = response_json["access_token"]
            self.__expires = datetime.today() + timedelta(
                seconds=response_json["expires_in"]
            )

            # Get the Modules as this now has moved to outside of the auth process
            url = urljoin(URL_BASE_NEW, URL_MODULES_NEW)
            response = await self.__session.get(
                url, headers={"Authorization": f"Bearer {self.__token}"}
            )

            # Response shoudl always be 200 even for login failures
            if response.status != 200:
                error_msg = await response.text()
                raise MasterthermConnectionError(str(response.status), error_msg)

            response_json = await response.json()

        # Next is same for old and new process, doing a double check just incase
        if response_json["returncode"] != 0:
            raise MasterthermAuthenticationError(
                response_json["returncode"], response_json["message"]
            )

        # Check if role is supported
        if not response_json["role"] in SUPPORTED_ROLES:
            raise MasterthermUnsupportedRole(
                "2", "Unsupported Role " + response_json["role"]
            )

        return response_json

    async def get_device_info(self, module_id: str, unit_id: str) -> dict:
        """Get the Device information.

        Parameters:
            module_id (str): This is the module_id for the unit
            unit_id (str): This is the unit id for the unit

        Return:
            device_info (dict): Information for a specific device.

        Raises:
            MasterthermConnectionError - General Connection Issue
            MasterthermTokenInvalid - Token has expired or is invalid
            MasterthermResponseFormatError - Some other issue, probably temporary"""
        params = f"moduleid={module_id}&unitid={unit_id}&application=android"

        try:
            response_json = await self.__get(
                url=URL_PUMPINFO if self.__api_version == "v1" else URL_PUMPINFO_NEW,
                params=params,
            )
        except MasterthermTokenInvalid:
            self.__expires = None
            response_json = await self.__get(
                url=URL_PUMPINFO if self.__api_version == "v1" else URL_PUMPINFO_NEW,
                params=params,
            )

        return response_json

    async def get_device_data(
        self, module_id: str, unit_id: str, last_update_time: str = None
    ) -> dict:
        """Get the Device lastest data.

        Parameters:
            module_id (str): This is the module_id for the unit
            unit_id (str): This is the unit id for the unit
            last_update_time (str): Optional last update date in number format

        Return:
            device_data (dict): data or updated data for a specific device.

        Raises:
            MasterthermConnectionError - General Connection Issue
            MasterthermTokenInvalid - Token has expired or is invalid
            MasterthermResponseFormatError - Some other issue, probably temporary"""
        params = f"moduleId={module_id}&deviceId={unit_id}&application=android&"
        if last_update_time is None:
            params = (
                params
                + "messageId=1&lastUpdateTime=0&errorResponse=true&fullRange=true"
            )
        else:
            params = (
                params
                + f"messageId=2&lastUpdateTime={last_update_time}&errorResponse=true&fullRange=true"
            )

        try:
            response_json = await self.__get(
                url=URL_PUMPDATA if self.__api_version == "v1" else URL_PUMPDATA_NEW,
                params=params,
            )
        except MasterthermTokenInvalid:
            self.__expires = None
            response_json = await self.__get(
                url=URL_PUMPDATA if self.__api_version == "v1" else URL_PUMPDATA_NEW,
                params=params,
            )

        if response_json["data"] != {}:
            data_file = ""
            if self.__api_version == "v1":
                if "varfile_mt1_config1" in response_json["data"]:
                    data_file = "varfile_mt1_config1"
                elif "varfile_mt1_config2" in response_json["data"]:
                    data_file = "varfile_mt1_config2"
            elif self.__api_version == "v2":
                data_file = "varFileData"

            response_json["data"]["varData"] = response_json["data"][data_file]
            del response_json["data"][data_file]

        return response_json

    async def set_device_data(
        self, module_id: str, unit_id: str, register: str, value: any
    ) -> bool:
        """Set device data a specific register to a specific value.

        Parameters:
            module_id (str): This is the module_id for the unit
            unit_id (str): This is the unit id for the unit
            register (str): The Register to update
            value (str|float): The value to set.

        Return:
           success (bool): return true if succes and false if not.

        Raises:
            MasterthermConnectionError - General Connection Issue
            MasterthermTokenInvalid - Token has expired or is invalid
            MasterthermResponseFormatError - Some other issue, probably temporary"""
        params = (
            f"moduleId={module_id}&deviceId={unit_id}&"
            + "configFile=varfile_mt1_config&messageId=1&errorResponse=true&"
            + f"variableId={register}&variableValue={value}&application=android"
        )

        try:
            response_json = await self.__post(
                url=URL_POSTUPDATE
                if self.__api_version == "v1"
                else URL_POSTUPDATE_NEW,
                params=params,
            )
        except MasterthermTokenInvalid:
            self.__expires = None
            response_json = await self.__get(
                url=URL_PUMPDATA if self.__api_version == "v1" else URL_PUMPDATA_NEW,
                params=params,
            )

        # Get the Value that is returned and set.
        if self.__api_version == "v1":
            data_file = ""
            if "varfile_mt1_config1" in response_json["data"]:
                data_file = "varfile_mt1_config1"
            elif "varfile_mt1_config2" in response_json["data"]:
                data_file = "varfile_mt1_config2"

            set_value = response_json["data"][data_file][str(unit_id).zfill(3)][
                register
            ]
        else:
            set_value = response_json["data"]["data"][str(unit_id).zfill(3)][register]

        return set_value == value



"""Main Program, for Testing Mainly"""
import argparse
import asyncio
import getpass
import json
import logging

from natsort import natsorted

from aiohttp import ClientSession

from masterthermconnect.__version__ import __version__
from masterthermconnect.api import MasterthermAPI
from masterthermconnect.controller import MasterthermController
from masterthermconnect.exceptions import (
    MasterthermAuthenticationError,
    MasterthermConnectionError,
    MasterthermUnsupportedRole,
    MasterthermEntryNotFound,
    MasterthermError,
)

_LOGGER: logging.Logger = logging.getLogger(__name__)


def guess_type(value: str) -> any:
    """Convert an Augument Type to a best guess format."""
    try:
        return int(value)
    except ValueError:
        try:
            return float(value)
        except ValueError:
            if value.lower() == "true":
                return True
            elif value.lower() == "false":
                return False
    return value


def get_arguments(argv) -> argparse.Namespace:
    """Read the Arguments passed in."""
    # formatter_class=argparse.MetavarTypeHelpFormatter,
    parser = argparse.ArgumentParser(
        prog="masterthermconnect",
        description=(
            "Python Mastertherm Connect API Module, used for debug purposes, "
            "allows you to get and set registers and other information for testing, "
            "use with caution!!!"
        ),
        epilog=(
            "DO NOT RUN THIS TOO FREQENTLY, IT IS POSSIBLE TO GET YOUR IP BLOCKED, "
            "I think new new API is sensitive to logging in too frequently."
        ),
    )
    parser.add_argument(
        "--version",
        action="version",
        version="Mastertherm Connect API Version: " + __version__,
        help="display the Mastertherm Connect API version",
    )

    # Sub Commands are get and set:
    subparsers = parser.add_subparsers(
        title="commands",
        description=(
            "Valid commands to access the API, use -h to get "
            "more help after the command for specific help."
        ),
        help="Retrieve and Send data to or from the API.",
    )

    # Get has opitions for user/ password/ api-ver/ sensitive info and pretty
    parser_get = subparsers.add_parser(
        "get", help="Read data from the API and Display it"
    )

    parser_get.set_defaults(command="get")
    parser_get.add_argument("--user", type=str, help="login user for Mastertherm")
    parser_get.add_argument(
        "--password", type=str, help="login password for Mastertherm"
    )
    parser_get.add_argument(
        "--api-ver",
        type=str,
        choices=["v1", "v2"],
        default="v1",
        help="API Version to use: Default: v1 (pre 2022), v2 (post 2022)",
    )
    parser_get.add_argument(
        "--hide-sensitive",
        action="store_true",
        help="Hide the actual sensitive information, "
        + "used when creating debug information for sharing.",
    )
    parser_get.add_argument(
        "--pretty",
        action="store_true",
        help="Prettify the Output in JSON Format.",
    )

    subparser_get = parser_get.add_subparsers(title="get commands", required=True)
    parser_getdevice = subparser_get.add_parser(
        "devices", help="All Devices List associated with the account."
    )
    parser_getdevice.set_defaults(subcommand="devices")

    parser_getdata = subparser_get.add_parser(
        "data",
        help="Normalized data for a speicif device, e.g. data 1234 1",
    )
    parser_getdata.set_defaults(subcommand="data")
    parser_getdata.add_argument("module_id", type=str, help="Module ID e.g. 1234")
    parser_getdata.add_argument("unit_id", type=str, help="Unit Id e.g. 1")

    parser_getreg = subparser_get.add_parser(
        "reg",
        help=(
            "Registers for a specific device, "
            "e.g. get reg 1234 1 A_101,A102 or reg 1234 1 all"
        ),
    )
    parser_getreg.set_defaults(subcommand="reg")
    parser_getreg.add_argument("module_id", type=str, help="Module ID e.g. 1234")
    parser_getreg.add_argument("unit_id", type=str, help="Unit Id e.g. 1")
    parser_getreg.add_argument(
        "reg",
        type=str,
        help="List Registers e.g. A_330 or A_330,A_331 or 'all' for everything.",
    )

    # Set has opitions for user/ password/ api-ver
    parser_set = subparsers.add_parser(
        "set", help="Set data to the API and check the Result"
    )

    parser_set.set_defaults(command="set")
    parser_set.add_argument("--user", type=str, help="login user for Mastertherm")
    parser_set.add_argument(
        "--password", type=str, help="login password for Mastertherm"
    )
    parser_set.add_argument(
        "--api-ver",
        type=str,
        choices=["v1", "v2"],
        default="v1",
        help="API Version to use: Default: v1 (pre 2022), v2 (post 2022)",
    )

    subparser_set = parser_set.add_subparsers(title="set commands", required=True)
    parser_setreg = subparser_set.add_parser(
        "reg",
        help="Registers for a specific device, e.g. reg 1234 1 D_3 0",
    )
    parser_setreg.set_defaults(subcommand="reg")
    parser_setreg.add_argument("module_id", type=str, help="Module ID e.g. 1234")
    parser_setreg.add_argument("unit_id", type=str, help="Unit Id e.g. 1")
    parser_setreg.add_argument("reg", type=str, help="Register to Set")
    parser_setreg.add_argument(
        "value", type=str, help="Value to Set, all values are set as strings"
    )

    parser_setdata = subparser_set.add_parser(
        "data",
        help=(
            "Data for a specific device, e.g. reg 1234 1 hp_power_state true"
            " or reg 1234 1 heating_circuits.hc0.on true "
            "Acceptable values are string, boolean, float or integer."
        ),
    )
    parser_setdata.set_defaults(subcommand="data")
    parser_setdata.add_argument("module_id", type=str, help="Module ID e.g. 1234")
    parser_setdata.add_argument("unit_id", type=str, help="Unit Id e.g. 1")
    parser_setdata.add_argument(
        "data", type=str, help="Data to Set using dot notation for parent child."
    )
    parser_setdata.add_argument(
        "value",
        type=guess_type,
        help="Value to Set, all values are string, boolean, float or integer.",
    )

    return parser.parse_args(argv)


async def connect(
    username: str, password: str, api_version: str, refresh: bool
) -> MasterthermController:
    """Setup and Connect to the Mastertherm Server."""
    # Login to the Server.
    session = ClientSession()
    controller = MasterthermController(
        username, password, session, api_version=api_version
    )

    try:
        await controller.connect()
        if refresh:
            await controller.refresh()

        return controller
    except MasterthermConnectionError as mte:
        _LOGGER.error("Connection Failed %s", mte.message)
    except MasterthermAuthenticationError as mte:
        _LOGGER.error("Authentication Failed %s", mte.message)
    except MasterthermUnsupportedRole as mte:
        _LOGGER.error("Unsupported Role %s", mte.message)
    except MasterthermError as mte:
        _LOGGER.error("Other Error %s", mte.message)
    finally:
        await session.close()

    return None


async def set_reg(
    username: str,
    password: str,
    api_version: str,
    module_id: str,
    unit_id: str,
    register: str,
    value: str,
) -> bool:
    """Setup and Connect to the MasterthermAPI."""
    session = ClientSession()

    success = False
    try:
        api = MasterthermAPI(username, password, session, api_version=api_version)
        await api.connect()

        success = await api.set_device_data(module_id, unit_id, register, value)
        if success:
            data = await api.get_device_data(module_id, unit_id)
            print(
                "Registration after Update: %s = %s",
                register,
                data["data"]["varData"][unit_id.zfill(3)][register],
            )
        else:
            _LOGGER.error("Failed to Set the Device Registry.")

    except MasterthermConnectionError as mte:
        _LOGGER.error("Connection Failed %s", mte.message)
    except MasterthermAuthenticationError as mte:
        _LOGGER.error("Authentication Failed %s", mte.message)
    except MasterthermUnsupportedRole as mte:
        _LOGGER.error("Unsupported Role %s", mte.message)
    except MasterthermError as mte:
        _LOGGER.error("Other Error %s", mte.message)
    finally:
        await session.close()

    return success


async def set_data(
    username: str,
    password: str,
    api_version: str,
    module_id: str,
    unit_id: str,
    data: str,
    value: any,
) -> bool:
    """Setup and Connect to the MasterthermController to set data."""
    session = ClientSession()

    success = False
    try:
        controller = MasterthermController(
            username, password, session, api_version=api_version
        )
        controller.set_refresh_rate(data_refresh_seconds=0)
        await controller.connect()
        await controller.refresh()

        success = await controller.set_device_data_item(module_id, unit_id, data, value)
        if success:
            await controller.refresh(full_load=True)
            updated_item = controller.get_device_data_item(module_id, unit_id, data)
            print(f"Data after Update: {data} = {updated_item}")
        else:
            _LOGGER.error("Failed to Set the Device Data.")

    except MasterthermConnectionError as mte:
        _LOGGER.error("Connection Failed %s", mte.message)
    except MasterthermAuthenticationError as mte:
        _LOGGER.error("Authentication Failed %s", mte.message)
    except MasterthermUnsupportedRole as mte:
        _LOGGER.error("Unsupported Role %s", mte.message)
    except MasterthermEntryNotFound as mte:
        _LOGGER.error("Entry not Found, Update Failed %s", mte.message)
    except MasterthermError as mte:
        _LOGGER.error("Other Error %s", mte.message)
    finally:
        await session.close()

    return success


async def set_command(login_user: str, login_pass: str, args) -> int:
    """Set Command to set data/registry."""
    if args.subcommand == "reg":
        if not await set_reg(
            login_user,
            login_pass,
            args.api_ver,
            args.module_id,
            args.unit_id,
            args.reg,
            args.value,
        ):
            _LOGGER.error("Set Register Command Failed, %s == %s", args.reg, args.value)
            return 2
    elif args.subcommand == "data":
        if not await set_data(
            login_user,
            login_pass,
            args.api_ver,
            args.module_id,
            args.unit_id,
            args.data,
            args.value,
        ):
            _LOGGER.error("Set Data Command Failed, %s == %s", args.data, args.value)
            return 2

    return 0


async def get_command(login_user: str, login_pass: str, args) -> int:
    """Get Command to get data/ registry/ devices."""
    controller = await connect(login_user, login_pass, args.api_ver, True)
    if controller is None:
        _LOGGER("Connection Failed.")
        return 1

    if args.subcommand == "devices":
        devices = controller.get_devices()
        new_module_id = 1111
        old_module_id = ""
        new_devices = {}

        for device_id, device_item in devices.items():
            module_id = device_item["module_id"]
            unit_id = device_item["unit_id"]

            if module_id != old_module_id:
                old_module_id = module_id
                new_module_id = new_module_id + 1

            if args.hide_sensitive:
                device_id = f"{str(new_module_id)}_{unit_id}"
                device_item["module_id"] = str(new_module_id)
                device_item["module_name"] = "Hidden Name"
                device_item["name"] = "First"
                device_item["surname"] = "Last"
                device_item["latitude"] = "1.1"
                device_item["longitude"] = "-0.1"

            new_devices[device_id] = device_item

        if args.pretty:
            print(json.dumps(new_devices, indent=4))
        else:
            print(str(new_devices).replace("'", '"'))
    elif args.subcommand == "data":
        module_id = args.module_id
        unit_id = args.unit_id
        device_data = controller.get_device_data(module_id, unit_id)
        if args.pretty:
            print(json.dumps(device_data, indent=4))
        else:
            print(str(device_data).replace("'", '"'))
    elif args.subcommand == "reg":
        module_id = args.module_id
        unit_id = args.unit_id
        device_reg = controller.get_device_registers(module_id, unit_id)

        sorted_reg = {}
        for key in natsorted(device_reg.keys()):
            sorted_reg[key] = device_reg[key]

        reg: str = args.reg
        if reg.upper() == "ALL":
            if args.pretty:
                print(json.dumps(sorted_reg, indent=4))
            else:
                print(str(sorted_reg).replace("'", '"'))
        elif reg.find(",") == -1:
            print(reg + " = " + sorted_reg.get(reg, "Not Found"))
        else:
            for key in reg.split(","):
                print(key + " = " + sorted_reg.get(key, "Not Found"))

    return 0


def main(argv=None) -> int:
    """Allow for testing connectivity from the command line."""
    # Arg Parse raises SystemExit, get return value
    try:
        args: argparse.Namespace = get_arguments(argv)
    except SystemExit as ex:
        return ex.code

    # If User/ Pass is not provided then get from the command line.
    login_user = input("User: ") if args.user is None else args.user
    login_pass = getpass.getpass() if args.password is None else args.password

    if args.command == "get":
        return asyncio.run(get_command(login_user, login_pass, args))
    elif args.command == "set":
        return asyncio.run(set_command(login_user, login_pass, args))

    return 0


if __name__ == "__main__":
    exit(main())



"""Constants used by the module."""
DATE_FORMAT = "%a, %d-%b-%Y %H:%M:%S %Z"
SUPPORTED_ROLES = ["400"]
SUPPORTED_API_VERSIONS = ["v1", "v2"]

# Old Mastertherm Devices Pre 2022
# This mostly uses POST methods
APP_CLIENTINFO = "os=android&osversion=7.0&ver=8&info=Google%2CAndroid"
URL_BASE = "https://mastertherm.vip-it.cz"
URL_LOGIN = "/plugins/mastertherm_login/client_login.php"
URL_PUMPINFO = "/plugins/get_pumpinfo/get_pumpinfo.php"
URL_PUMPDATA = "/mt/PassiveVizualizationServlet"
URL_POSTUPDATE = "/mt/ActiveVizualizationServlet"

# New Mastertherm Devices (Mastertherm Touch) Post 2022
# This uses GET to retrieve data and POST to submit, e.g. Login is Post, Update is Post
APP_CLIENTINFO_NEW = "client_id=mobile-android"
URL_BASE_NEW = "https://mastertherm.online"
URL_LOGIN_NEW = "/auth/realms/neobox/protocol/openid-connect/token"
URL_PUMPINFO_NEW = "/api/v1/hp_info"
URL_PUMPDATA_NEW = "/api/v1/hp_data"
URL_MODULES_NEW = "/api/v1/modules"
URL_POSTUPDATE_NEW = "/api/v1/hp_data"

# Used to setup the hc name from the registers,
# for some reason they don't like the letter Q
CHAR_MAP = [
    "-",
    "A",
    "B",
    "C",
    "D",
    "E",
    "F",
    "G",
    "H",
    "I",
    "J",
    "K",
    "L",
    "M",
    "N",
    "O",
    "P",
    "R",
    "S",
    "T",
    "U",
    "V",
    "W",
    "X",
    "Y",
    "Z",
]

# HC0 is the main heating cooling circuit which is enabled if HC1 to HC6 are not used.
# HC1 to HC6 are optional heating and cooling circuits that may be installed
# Pool and Solar are spencial circuit
HC_MAP = {
    0: {"id": "hc0", "pad": "padz", "register": "D_182", "default": "Home"},
    1: {"id": "hc1", "pad": "pada", "register": "D_278", "default": ""},
    2: {"id": "hc2", "pad": "padb", "register": "D_436", "default": ""},
    3: {"id": "hc3", "pad": "padc", "register": "D_298", "default": ""},
    4: {"id": "hc4", "pad": "padd", "register": "D_307", "default": ""},
    5: {"id": "hc5", "pad": "pade", "register": "D_316", "default": ""},
    6: {"id": "hc6", "pad": "padf", "register": "D_326", "default": ""},
}

DEVICE_INFO_MAP = {
    "name": "givenname",
    "surname": "surname",
    "country": "localization",
    "language": "lang",
    "hp_type": "type",
    "controller": "regulation",
    "exp": "exp",
    "output": "output",
    "cooling": "reversation",  # 0 - Cooling Disabled, 1 - Cooling Enabled.
    "place": "city",
    "latitude": "password9",
    "longitude": "password10",
    "notes": "notes",
    "pada": "pada",
    "padb": "padb",
    "padc": "padc",
    "padd": "padd",
    "pade": "pade",
    "padf": "padf",
    "padz": "padz",
}



"""Mapping files for reading from the registries."""
from .special import Special

# Heating and Cooling circuits, hc0 is the default
# hc0 gets disabled if optional hc1 to hc6 are installed
# Pool and Solar have been added to the heating cooling circuits
#
# PAD
#   entries are related to the Room Thermostats
#   Active is False if not installed, so will disable the section
DEVICE_READ_HCMAP = {
    "hc0": {
        "enabled": [Special(bool, Special.FIXED), False],
        "name": [Special(str, Special.NAMEARRAY), []],  # hc0 does not have a name
        "ambient_requested": [
            Special(float, Special.FORMULA),
            [
                "{1} if {0} else {2}",
                [[bool, "D_242"], [float, "A_191"], [float, "A_210"]],
            ],
        ],
        "ambient_temp": [float, "A_211"],
        "pad": {
            "enabled": [
                Special(bool, Special.FORMULA),
                ["not {0}", [[bool, "D_242"]]],
            ],  # Seems to be enabled if this is false.
            "current_humidity": [float, "I_185"],
        },
    },
    "hc1": {
        "enabled": [Special(bool, Special.FIXED), False],
        "name": [
            Special(str, Special.NAMEARRAY),
            ["I_211", "I_212", "I_213", "I_214", "I_215", "I_216"],
        ],
        "on": [bool, "D_212"],
        "cooling": [bool, "D_213"],
        "circulation_valve": [bool, "D_68"],
        "water_temp": [float, "A_90"],
        "water_requested": [float, "A_96"],
        "auto": [int, "I_269"],
        "ambient_requested": [
            Special(float, Special.FORMULA),
            [
                "{1} if {0} else {2}",
                [[bool, "D_245"], [float, "A_219"], [float, "A_215"]],
            ],
        ],
        "ambient_temp": [float, "A_216"],
        "pad": {
            "enabled": [bool, "D_245"],
            "state": [int, "I_15"],  # 0 - Permanently Off, 1 - Scheduled Off, 2 - On
            "current_humidity": [float, "I_219"],
        },
        "control_curve_heating": {
            "setpoint_a_outside": [float, "A_101"],
            "setpoint_a_requested": [float, "A_106"],
            "setpoint_b_outside": [float, "A_102"],
            "setpoint_b_requested": [float, "A_107"],
        },
        "control_curve_cooling": {
            "setpoint_a_outside": [float, "A_314"],
            "setpoint_a_requested": [float, "A_315"],
            "setpoint_b_outside": [float, "A_316"],
            "setpoint_b_requested": [float, "A_317"],
        },
    },
    "hc2": {
        "enabled": [Special(bool, Special.FIXED), False],
        "name": [
            Special(str, Special.NAMEARRAY),
            ["I_221", "I_222", "I_223", "I_224", "I_225", "I_226"],
        ],
        "on": [bool, "D_216"],
        "cooling": [bool, "D_217"],
        "circulation_valve": [bool, "D_69"],
        "water_temp": [float, "A_91"],
        "water_requested": [float, "A_97"],
        "auto": [int, "I_270"],
        "ambient_requested": [
            Special(float, Special.FORMULA),
            [
                "{1} if {0} else {2}",
                [[bool, "D_248"], [float, "A_225"], [float, "A_221"]],
            ],
        ],
        "ambient_temp": [float, "A_222"],
        "pad": {
            "enabled": [bool, "D_248"],
            "current_humidity": [float, "I_220"],
        },
        "control_curve_heating": {
            "setpoint_a_outside": [float, "A_108"],
            "setpoint_a_requested": [float, "A_84"],
            "setpoint_b_outside": [float, "A_109"],
            "setpoint_b_requested": [float, "A_85"],
        },
        "control_curve_cooling": {
            "setpoint_a_outside": [float, "A_330"],
            "setpoint_a_requested": [float, "A_331"],
            "setpoint_b_outside": [float, "A_332"],
            "setpoint_b_requested": [float, "A_333"],
        },
    },
    "hc3": {
        "enabled": [Special(bool, Special.FIXED), False],
        "name": [
            Special(str, Special.NAMEARRAY),
            ["I_231", "I_232", "I_233", "I_234", "I_235", "I_236"],
        ],
        "on": [bool, "D_220"],
        "cooling": [bool, "D_221"],
        "circulation_valve": [bool, "D_70"],
        "water_temp": [float, "A_92"],
        "water_requested": [float, "A_98"],
        "auto": [int, "I_271"],
        "ambient_requested": [
            Special(float, Special.FORMULA),
            [
                "{1} if {0} else {2}",
                [[bool, "D_251"], [float, "A_231"], [float, "A_227"]],
            ],
        ],
        "ambient_temp": [float, "A_228"],
        "pad": {
            "enabled": [bool, "D_251"],
            "current_humidity": [float, "I_221"],
        },
        "control_curve_heating": {
            "setpoint_a_outside": [float, "A_113"],
            "setpoint_a_requested": [float, "A_86"],
            "setpoint_b_outside": [float, "A_114"],
            "setpoint_b_requested": [float, "A_87"],
        },
        "control_curve_cooling": {
            "setpoint_a_outside": [float, "A_346"],
            "setpoint_a_requested": [float, "A_347"],
            "setpoint_b_outside": [float, "A_348"],
            "setpoint_b_requested": [float, "A_349"],
        },
    },
    "hc4": {
        "enabled": [Special(bool, Special.FIXED), False],
        "name": [
            Special(str, Special.NAMEARRAY),
            ["I_241", "I_242", "I_243", "I_244", "I_245", "I_246"],
        ],
        "on": [bool, "D_50"],
        "cooling": [bool, "D_224"],
        "circulation_valve": [bool, "D_71"],
        "water_temp": [float, "A_93"],
        "water_requested": [float, "A_99"],
        "auto": [int, "I_272"],
        "current_humidity": [float, "I_222"],
        "ambient_requested": [
            Special(float, Special.FORMULA),
            [
                "{1} if {0} else {2}",
                [[bool, "D_254"], [float, "A_238"], [float, "A_233"]],
            ],
        ],
        "ambient_temp": [float, "A_234"],
        "pad": {
            "enabled": [bool, "D_254"],
            "current_humidity": [float, "I_222"],
        },
        "control_curve_heating": {
            "setpoint_a_outside": [float, "A_122"],
            "setpoint_a_requested": [float, "A_120"],
            "setpoint_b_outside": [float, "A_88"],
            "setpoint_b_requested": [float, "A_121"],
        },
        "control_curve_cooling": {
            "setpoint_a_outside": [float, "A_362"],
            "setpoint_a_requested": [float, "A_363"],
            "setpoint_b_outside": [float, "A_364"],
            "setpoint_b_requested": [float, "A_365"],
        },
    },
    "hc5": {
        "enabled": [Special(bool, Special.FIXED), False],
        "name": [
            Special(str, Special.NAMEARRAY),
            ["I_251", "I_252", "I_253", "I_254", "I_255", "I_256"],
        ],
        "on": [bool, "D_51"],
        "cooling": [bool, "D_227"],
        "circulation_valve": [bool, "D_72"],
        "water_temp": [float, "A_243"],
        "water_requested": [float, "A_242"],
        "auto": [int, "I_273"],
        "ambient_requested": [
            Special(float, Special.FORMULA),
            [
                "{1} if {0} else {2}",
                [[bool, "D_257"], [float, "A_247"], [float, "A_240"]],
            ],
        ],
        "ambient_temp": [float, "A_241"],
        "pad": {
            "enabled": [bool, "D_257"],
            "current_humidity": [float, "I_223"],
        },
        "control_curve_heating": {
            "setpoint_a_outside": [float, "A_387"],
            "setpoint_a_requested": [float, "A_388"],
            "setpoint_b_outside": [float, "A_389"],
            "setpoint_b_requested": [float, "A_390"],
        },
        "control_curve_cooling": {
            "setpoint_a_outside": [float, "A_379"],
            "setpoint_a_requested": [float, "A_380"],
            "setpoint_b_outside": [float, "A_381"],
            "setpoint_b_requested": [float, "A_382"],
        },
    },
    "hc6": {
        "enabled": [Special(bool, Special.FIXED), False],
        "name": [
            Special(str, Special.NAMEARRAY),
            ["I_261", "I_262", "I_263", "I_264", "I_265", "I_266"],
        ],
        "on": [bool, "D_52"],
        "cooling": [bool, "D_231"],
        "circulation_valve": [bool, "D_73"],
        "water_temp": [float, "A_252"],
        "water_requested": [float, "A_251"],
        "auto": [int, "I_274"],
        "current_humidity": [float, "I_224"],
        "ambient_requested": [
            Special(float, Special.FORMULA),
            [
                "{1} if {0} else {2}",
                [[bool, "D_259"], [float, "A_249"], [float, "A_277"]],
            ],
        ],
        "ambient_temp": [float, "A_250"],
        "pad": {
            "enabled": [bool, "D_259"],
            "current_humidity": [float, "I_224"],
        },
        "control_curve_heating": {
            "setpoint_a_outside": [float, "A_401"],
            "setpoint_a_requested": [float, "A_402"],
            "setpoint_b_outside": [float, "A_403"],
            "setpoint_b_requested": [float, "A_404"],
        },
        "control_curve_cooling": {
            "setpoint_a_outside": [float, "A_405"],
            "setpoint_a_requested": [float, "A_406"],
            "setpoint_b_outside": [float, "A_407"],
            "setpoint_b_requested": [float, "A_408"],
        },
    },
    "solar": {
        "enabled": [bool, "D_433"],
        "name": [Special(str, Special.FIXED), "Solar"],
        "solar_collector": [float, "A_258"],
        "water_tank1": [float, "A_259"],
        "water_tank2": [float, "A_260"],
    },
    "pool": {
        "name": [Special(str, Special.FIXED), "Pool"],
        "enabled": [bool, "D_348"],
        "on": [bool, "D_238"],
        "heating": [bool, "D_43"],
        "temp_actual": [float, "A_262"],
        "temp_requested": [float, "A_210"],
    },
}

# NOTES --------------------------------------------------
# Moved HC0 Heating/ Cooling Curves to main as always
# used moving to main section so they don't get disabled
# --------------------------------------------------------
DEVICE_READ_MAP = {
    "hp_power_state": [bool, "D_3"],
    "hp_function": [int, "I_51"],  # 0: heating, #1: cooling, #2: auto (Write)
    "operating_mode": [
        Special(str, Special.FIXED),
        "heating",
    ],  # heating, cooling, pool, dhw, dpc
    "season": {
        "mode": [
            Special(str, Special.FORMULA),
            [
                "('' if {0} else 'auto-') + ('winter' if {1} else 'summer')",
                [[bool, "I_50"], [bool, "D_24"]],
            ],
        ],
        "manual_set": [bool, "I_50"],
        "winter": [bool, "D_24"],
        "winter_temp": [float, "A_82"],
        "summer_temp": [float, "A_83"],
    },
    "cooling_mode": [bool, "D_4"],
    "control_curve_heating": {
        "setpoint_a_outside": [float, "A_35"],
        "setpoint_a_requested": [float, "A_37"],
        "setpoint_b_outside": [float, "A_36"],
        "setpoint_b_requested": [float, "A_38"],
    },
    "control_curve_cooling": {
        "setpoint_a_outside": [float, "A_47"],
        "setpoint_a_requested": [float, "A_48"],
        "setpoint_b_outside": [float, "A_49"],
        "setpoint_b_requested": [float, "A_50"],
    },
    "domestic_hot_water": {
        "heating": [bool, "D_66"],
        "enabled": [bool, "D_275"],
        "current_temp": [float, "A_126"],
        "required_temp": [float, "A_129"],
        "min_temp": [float, "A_296"],
        "max_temp": [float, "A_297"],
    },
    "compressor_running": [bool, "D_5"],
    "compressor2_running": [bool, "D_32"],
    "circulation_pump_running": [bool, "D_10"],
    "fan_running": [bool, "D_8"],
    "defrost_mode": [bool, "D_11"],
    "aux_heater_1": [bool, "D_6"],
    "aux_heater_2": [bool, "D_7"],
    "outside_temp": [float, "A_3"],
    "requested_temp": [float, "A_5"],
    "actual_temp": [float, "A_1"],
    "dewp_control": [bool, "D_196"],
    "hdo_on": [bool, "D_15"],
    "runtime_info": {
        "compressor_run_time": [int, "I_11"],
        "compressor_start_counter": [int, "I_12"],
        "pump_runtime": [int, "I_13"],
        "aux1_runtime": [int, "I_100"],
        "aux2_runtime": [int, "I_101"],
    },
    "error_info": {
        "some_error": [bool, "D_20"],
        "three_errors": [bool, "D_21"],
        "reset_3e": [bool, "D_19"],
        "safety_tstat": [bool, "D_77"],
        "alarm_a": [int, "I_20"],
        "alarm_b": [int, "I_21"],
        "alarm_c": [int, "I_22"],
        "alarm_d": [int, "I_23"],
        "alarm_e": [int, "I_24"],
        "alarm_f": [int, "I_25"],
        "alarm_g": [int, "I_26"],
        "alarm_h": [int, "I_27"],
        "alarm_i": [int, "I_28"],
        "alarm_j": [int, "I_39"],
        "alarm_k": [int, "I_40"],
        "alarm_l": [int, "I_41"],
        "alarm_m": [int, "I_42"],
    },
    "heating_circuits": DEVICE_READ_HCMAP,
}



"""Python API wrapper for Mastertherm Connect."""
from masterthermconnect.__version__ import __version__
from masterthermconnect.controller import MasterthermController
from masterthermconnect.exceptions import (
    MasterthermAuthenticationError,
    MasterthermConnectionError,
    MasterthermResponseFormatError,
    MasterthermTokenInvalid,
    MasterthermUnsupportedRole,
    MasterthermUnsupportedVersion,
    MasterthermEntryNotFound,
)

__all__ = [
    "__version__",
    "MasterthermController",
    "MasterthermAuthenticationError",
    "MasterthermConnectionError",
    "MasterthermResponseFormatError",
    "MasterthermTokenInvalid",
    "MasterthermUnsupportedRole",
    "MasterthermUnsupportedVersion",
]
